<?php

declare(strict_types=1);

namespace Drupal\role_revocation\Commands;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\role_revocation\RevocationHelper;
use Drush\Commands\DrushCommands;

/**
 * Revoke Roles Drush commands.
 */
class RevokeRolesDrushCommands extends DrushCommands {

  /**
   * User storage manager.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected $userStorage;

  /**
   * Helper for fetching data.
   *
   * @var \Drupal\role_revocation\RevocationHelper
   */
  protected $revocationHelper;

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * The date formatter service.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * Constructs a new RevokeRolesDrushCommands object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   * @param \Drupal\role_revocation\RevocationHelper $revocationHelper
   *   The revocation helper service.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The configuration factory service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\Core\Datetime\DateFormatterInterface $dateFormatter
   *   The date formatter service.
   */
  public function __construct(EntityTypeManagerInterface $entityTypeManager, RevocationHelper $revocationHelper, ConfigFactoryInterface $configFactory, TimeInterface $time, DateFormatterInterface $dateFormatter) {
    $this->userStorage = $entityTypeManager->getStorage('user');
    $this->revocationHelper = $revocationHelper;
    $this->configFactory = $configFactory;
    $this->time = $time;
    $this->dateFormatter = $dateFormatter;
  }

  /**
   * Command to revoke roles from inactive editors.
   *
   * @param array $options
   *   Optional dry run.
   *
   * @command role_revocation:revoke
   * @options dry Indicate this is a dry run.
   * @usage role_revocation:revoke --dry
   *   Simulates the role revocation, but only do a dry run.
   */
  public function revokeRoles(array $options = [
    'dry' => FALSE,
  ]): void {
    $isDry = !empty($options['dry']);

    $userCount = 0;

    $config = $this->configFactory->get('role_revocation.settings');
    $revokeThresholdDays = $config->get('threshold') ?? 365;
    $revocationThreshold = $this->time->getRequestTime() - ($revokeThresholdDays * 86400);

    $privilegedRoles = $config->get('privileged_roles') ?? [];

    $revocableRoles = array_diff($config->get('roles') ?? [], $privilegedRoles);

    if (empty($revocableRoles)) {
      return;
    }

    $usersToRevoke = $this->revocationHelper->getUsersAndRolesToBeRevoked($revocableRoles, $revocationThreshold, $privilegedRoles);

    foreach ($usersToRevoke as $userInfo) {
      $user = $userInfo['user'];
      $removeRoles = $userInfo['roles'];
      $roles = $user->getRoles(TRUE);

      $userCount++;

      $newRoles = array_diff($roles, $removeRoles);

      if ($isDry) {
        $this->output()->writeln(sprintf('Following roles would be removed for user %s (%s): (%s)', $user->label(), $user->get('mail')->value, implode(', ', $removeRoles)));
        continue;
      }

      $user->set('roles', $newRoles);
      $user->save();
    }

    if (!$isDry) {
      $this->output()->writeln(sprintf('Revoked roles for %s users.', $userCount));
    }
  }

}
