<?php

declare(strict_types=1);

namespace Drupal\role_revocation\Form;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\Datetime\DateFormatterInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\role_revocation\RevocationHelper;
use Drupal\user\UserStorageInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Listing of role revocations.
 *
 * @package Drupal\role_revocation\Form
 */
class RevocationsForm extends FormBase {

  /**
   * The configuration factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * User storage manager.
   *
   * @var \Drupal\user\UserStorageInterface
   */
  protected $userStorage;

  /**
   * Helper for fetching data.
   *
   * @var \Drupal\role_revocation\RevocationHelper
   */
  protected $revocationHelper;

  /**
   * Date formatter.
   *
   * @var \Drupal\Core\Datetime\DateFormatterInterface
   */
  protected $dateFormatter;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'role_revocation_revocations';
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(Connection $database, UserStorageInterface $userStorage, RevocationHelper $revocationHelper, ConfigFactoryInterface $configFactory, TimeInterface $time, DateFormatterInterface $dateFormatter) {
    $this->dateFormatter = $dateFormatter;
    $this->database = $database;
    $this->userStorage = $userStorage;
    $this->revocationHelper = $revocationHelper;
    $this->configFactory = $configFactory;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('entity_type.manager')->getStorage('user'),
      $container->get('role_revocation.revocation_helper'),
      $container->get('config.factory'),
      $container->get('datetime.time'),
      $container->get('date.formatter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $activeUids = [];

    $config = $this->configFactory->get('role_revocation.settings');

    $revocationthresholdDays = $config->get('threshold') ?? 365;
    $revocationThreshold = $this->time->getRequestTime() - ($revocationthresholdDays * 86400);

    // Calculate the 10% threshold for upcoming revocations.
    $percentThresholdDays = $revocationthresholdDays * 0.1;
    $percentRevocationThreshold = strtotime("+$percentThresholdDays days", $revocationThreshold);

    $privilegedRoles = $config->get('privileged_roles');

    $recentlyActive = $this->revocationHelper->getRecentlyActiveUserData($revocationThreshold);
    $recentlyInactive = $this->revocationHelper->getRecentlyInactiveUserData($revocationThreshold);

    $activeUids = array_flip($activeUids);

    $revocableRoles = array_diff($config->get('roles') ?? [], $config->get('privileged_roles') ?? []);

    $userResult = $this->revocationHelper->getUsersWithRevocableRoles($revocableRoles, $revocationThreshold);

    $inactiveRows = [];
    $upcomingRevocationRows = [];

    foreach ($userResult as $row) {
      $user = $this->userStorage->load($row->entity_id);

      // Skip privileged roles.
      $roles = $user->getRoles(TRUE);
      if (array_intersect($roles, $privilegedRoles)) {
        continue;
      }

      // Check if the roles were assigned before the revocation threshold.
      $roleRevocations = $this->revocationHelper->getRoleRevocations($user->id());
      $rolesToBeRevoked = [];
      foreach ($roles as $role) {
        // Check if the role is set to be revoked and if it was assigned before
        // the threshold.
        if (isset($roleRevocations[$role]) && $roleRevocations[$role] < $revocationThreshold && in_array($role, $revocableRoles)) {
          $rolesToBeRevoked[] = $role;
        }
      }

      if (empty($rolesToBeRevoked)) {
        continue;
      }

      $roleAssignmentDates = array_map(function ($role) use ($roleRevocations) {
        return $role . ': ' . $this->dateFormatter->format($roleRevocations[$role], 'custom', 'Y-m-d H:i');
      }, $rolesToBeRevoked);

      if (!isset($recentlyActive[$row->entity_id])) {
        $contentChanged = (isset($recentlyInactive[$user->id()]) ?
          $this->dateFormatter->format($recentlyInactive[$user->id()]->timestamp, 'custom', 'Y-m-d H:i') :
          $this->t('never'));
        $inactiveRows[] = [
          $user->toLink(),
          ['data' => ['#markup' => implode('<br />', $rolesToBeRevoked)]],
          $contentChanged,
          ['data' => ['#markup' => implode('<br />', $roleAssignmentDates)]],
        ];
      }
      elseif ($recentlyActive[$row->entity_id]->timestamp < $percentRevocationThreshold) {
        $upcomingRevocationRows[] = [
          $user->toLink(),
          ['data' => ['#markup' => implode('<br />', $rolesToBeRevoked)]],
          $this->dateFormatter->format($recentlyActive[$user->id()]->timestamp, 'custom', 'Y-m-d H:i'),
          ['data' => ['#markup' => implode('<br />', $roleAssignmentDates)]],
        ];
      }
    }

    $header = [
      $this->t('User'),
      $this->t('Roles'),
      $this->t('Latest edit of content'),
      $this->t('Role assignment dates'),
    ];

    $form['inactiveList'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $inactiveRows,
      '#empty' => $this->t('No users found'),
      '#caption' => $this->t('Role revocation will happen on next cron run (@count users)', ['@count' => count($inactiveRows)]),
    ];

    $form['list'] = [
      '#type' => 'table',
      '#header' => $header,
      '#rows' => $upcomingRevocationRows,
      '#empty' => $this->t('No users found'),
      '#caption' => $this->t('Role revocation happening soon (@count users)', ['@count' => count($upcomingRevocationRows)]),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {

  }

}
