<?php

declare(strict_types=1);

namespace Drupal\role_revocation;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Database\Connection;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Revocation helper.
 *
 * Provides functions for querying the database in order to determine users
 * whose roles should be revoked.
 */
final class RevocationHelper implements ContainerInjectionInterface {

  /**
   * The database connection.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * {@inheritdoc}
   */
  public function __construct(Connection $database, EntityTypeManagerInterface $entityTypeManager, TimeInterface $time) {
    $this->database = $database;
    $this->entityTypeManager = $entityTypeManager;
    $this->time = $time;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('database'),
      $container->get('entity_type.manager'),
      $container->get('datetime.time')
    );
  }

  /**
   * Get recently active users.
   *
   * @param int $timestamp
   *   The threshold as Unix timestamp to consider for recent activity.
   *
   * @return array
   *   An array of results.
   */
  public function getRecentlyActiveUserData(int $timestamp): array {
    $query = $this->database->query('SELECT nr.revision_uid, MAX(nr.revision_timestamp) AS timestamp, u.name FROM {node_revision} nr, {users_field_data} u WHERE nr.revision_uid = u.uid AND nr.revision_uid != 0 AND revision_timestamp > :timestamp GROUP BY nr.revision_uid, u.name ORDER BY timestamp ASC', [
      ':timestamp' => $timestamp,
    ]);

    $result = $query->fetchAllAssoc('revision_uid');

    return $result;
  }

  /**
   * Get recently inactive users. Currently the default is defined as 1 year.
   *
   * @param int $inactiveSince
   *   Timestamp defining the moment user is considered inactive.
   *
   * @return array
   *   An array of results.
   */
  public function getRecentlyInactiveUserData(int $inactiveSince = NULL): array {
    $query = $this->database->query('SELECT nr.revision_uid, MAX(nr.revision_timestamp) AS timestamp, u.name FROM {node_revision} nr, {users_field_data} u WHERE nr.revision_uid = u.uid AND nr.revision_uid != 0 AND revision_timestamp < :timestamp GROUP BY nr.revision_uid, u.name ORDER BY timestamp ASC', [
      ':timestamp' => $inactiveSince ?? strtotime('-1 year'),
    ]);

    $result = $query->fetchAllAssoc('revision_uid');

    return $result;
  }

  /**
   * Get users with roles that are affected by revocation.
   *
   * @param array $revocableRoles
   *   An array of roles that are subject to revocation.
   * @param int $threshold
   *   The timestamp threshold for revocation.
   *
   * @return array
   *   An array of results.
   */
  public function getUsersWithRevocableRoles(array $revocableRoles, int $threshold): array {
    if (empty($revocableRoles)) {
      return [];
    }
    $placeholders = [];
    foreach ($revocableRoles as $key => $role) {
      $placeholders[":role_$key"] = $role;
    }

    $placeholders[':threshold'] = $threshold;
    $roles_in = implode(', ', array_keys($placeholders));

    $query = $this->database->query(
      "SELECT ur.entity_id, COUNT(ur.roles_target_id) AS role_count
       FROM {user__roles} ur
       JOIN {users_field_data} u ON ur.entity_id = u.uid
       JOIN {role_revocation} rr ON ur.entity_id = rr.uid AND ur.roles_target_id = rr.role
       WHERE ur.roles_target_id IN ($roles_in) AND rr.assigned < :threshold
       GROUP BY ur.entity_id",
      $placeholders
    );

    $result = $query->fetchAll();

    return $result;
  }

  /**
   * Gets role revocations for a user.
   *
   * @param int $uid
   *   The user ID.
   *
   * @return array
   *   An array of roles and their assigned timestamps.
   */
  public function getRoleRevocations($uid) {
    $query = $this->database->select('role_revocation', 'rr')
      ->fields('rr', ['role', 'assigned'])
      ->condition('uid', $uid);
    $results = $query->execute()->fetchAllKeyed();

    return $results;
  }

  /**
   * Tracks role assignments with a timestamp.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being saved.
   */
  public function trackRoleAssignments(EntityInterface $entity) {
    // Load the original user entity from the database to compare.
    $originalAccount = $entity->original ?? NULL;
    $originalRoles = $originalAccount ? $originalAccount->getRoles() : [];

    $currentRoles = $entity->getRoles();

    $newRoles = array_diff($currentRoles, $originalRoles);

    $removedRoles = array_diff($originalRoles, $currentRoles);

    $timestamp = $this->time->getRequestTime();

    // Insert or update role assignments for new roles.
    foreach ($newRoles as $role) {
      $this->database->merge('role_revocation')
        ->key([
          'uid' => $entity->id(),
          'role' => $role,
        ])
        ->fields([
          'assigned' => $timestamp,
        ])
        ->execute();
    }

    // Update role assignments for removed roles.
    foreach ($removedRoles as $role) {
      $this->database->delete('role_revocation')
        ->condition('uid', $entity->id())
        ->condition('role', $role)
        ->execute();
    }
  }

  /**
   * Removes all role assignment information for specific user entity.
   *
   * @param \Drupal\user\UserInterface $entity
   *   User entity to remove assignment information from.
   */
  public function removeAllRoleAssignments($entity) {
    if (!$entity instanceof UserInterface) {
      return;
    }
    $this->database->delete('role_revocation')
      ->condition('uid', $entity->id())
      ->execute();
  }

  /**
   * Get users with revocable roles and roles to be revoked.
   *
   * @param array $revocableRoles
   *   An array of roles that are subject to revocation.
   * @param int $revocationThreshold
   *   The timestamp threshold for revocation.
   * @param array $privilegedRoles
   *   An array of privileged roles.
   *
   * @return array
   *   An array of user objects and roles to be revoked.
   */
  public function getUsersAndRolesToBeRevoked(array $revocableRoles, int $revocationThreshold, array $privilegedRoles): array {
    $activeUsers = $this->getRecentlyActiveUserData($revocationThreshold);
    $activeUids = array_flip(array_keys($activeUsers));
    $userResult = $this->getUsersWithRevocableRoles($revocableRoles, $revocationThreshold);

    $usersToRevoke = [];
    foreach ($userResult as $row) {
      if (isset($activeUids[$row->entity_id])) {
        continue;
      }

      $user = $this->entityTypeManager->getStorage('user')->load($row->entity_id);

      // Skip privileged roles.
      $roles = $user->getRoles(TRUE);
      if (array_intersect($roles, $privilegedRoles)) {
        continue;
      }

      // Check if the roles were assigned before the revocation threshold.
      $roleRevocations = $this->getRoleRevocations($user->id());
      $removeRoles = [];
      foreach ($roles as $role) {
        // Check if the role is set to be revoked and if it was assigned
        // before the threshold.
        if (isset($roleRevocations[$role]) && $roleRevocations[$role] < $revocationThreshold && in_array($role, $revocableRoles)) {
          $removeRoles[] = $role;
        }
      }

      if (!empty($removeRoles)) {
        $usersToRevoke[] = ['user' => $user, 'roles' => $removeRoles];
      }
    }

    return $usersToRevoke;
  }

}
